#!/usr/bin/env python

# (c) 2005 Billy Charlton
# Licensed under GNU GPL.
# See www.gnu.org for details.

import sys
import os
import popen2
import time
import random
import urllib
import ifp

import pygtk
pygtk.require("2.0")

import gtk
import gtk.glade
import gnome.ui

class GiRiverApp:
	def __init__(self):
		gladefile="ifp-gnome.glade"
		windowname="window1"
		progresswinname="winProgress"

		self.uploadSize = 0
		self.doneSoFarSize = 0
		
		self.wTree=gtk.glade.XML(gladefile,windowname)
		self.progTree=gtk.glade.XML(gladefile,progresswinname)

		dic = { "on_window1_destroy"    : (gtk.main_quit) ,
			"on_btnRefresh_clicked" : (self.refreshClicked),
			"on_btnClose_clicked" : (gtk.main_quit),
			"on_btnDelete_clicked" : (self.deleteClicked),
			"on_about1_activate" : (self.showAboutDialog),
			"on_btnAddPlaylist_clicked" : (self.addPlaylistClicked),
			"on_quit1_activate" : (gtk.main_quit),
			"on_add_files1_activate" : (self.addClicked),
			"on_upload_to_pc_clicked" : (self.copyToPC),
			"on_btnAdd_clicked" : (self.addClicked)
		}

		self.wTree.signal_autoconnect(dic)
		self.mainWindow = self.wTree.get_widget(windowname)
		self.mainWindow.hide()
		self.mainWindow.set_icon_from_file('ifp-gnome.png')

		# Progress Window
		self.progressWindow = self.progTree.get_widget(progresswinname)
		dictp = { 'on_btnProgressCancel_clicked': (self.cancelTransfer) }
		self.progTree.signal_autoconnect(dictp)

		
		self.progressWindow.hide()
		self.progressWindow.set_icon_from_file('ifp-gnome.png')

		# Transfer Window
		self.interrupted = False
		self.verDialog = gtk.glade.XML(gladefile,"winTransfer")
		dictt = {
			"on_btnDownload_clicked" : (self.transferClicked),
			"on_btnCancel_clicked"   : (self.transferCancelled)
		}
		self.verDialog.signal_autoconnect(dictt)
		self.verifyWindow = self.verDialog.get_widget("winTransfer")
		self.verifyWindow.hide()
		self.verifyWindow.set_icon_from_file('ifp-gnome.png')

		self.lblBattery = self.wTree.get_widget("lblBattery")
		self.lblSpace = self.wTree.get_widget("lblSpace")

		self.tree = self.wTree.get_widget("treeview1")
		self.progBar = self.progTree.get_widget("progressbar1")
		self.progCancel = self.progTree.get_widget("btnProgressCancel")

		while gtk.events_pending(): 
			gtk.main_iteration(False)
		
		#Filetree
		renderer = gtk.CellRendererText()
		column = gtk.TreeViewColumn('Name',renderer,text=0)
		column.set_clickable(True)
		column.set_sizing(gtk.TREE_VIEW_COLUMN_AUTOSIZE)
		column.set_property("resizable",True)
		column.set_expand(True)
		self.tree.append_column(column)

		#Filesize
		sizerenderer = gtk.CellRendererText()
		sizerenderer.set_property("xalign",1.0)
		sizeColumn = gtk.TreeViewColumn('Size',sizerenderer,text=1)
		sizeColumn.set_clickable(True)
		sizeColumn.set_sizing(gtk.TREE_VIEW_COLUMN_AUTOSIZE)
		sizeColumn.set_expand(False)
		self.tree.append_column(sizeColumn)

		self.setupDragAndDrop()

		self.contents = gtk.TreeStore(str, str)
		self.tree.set_model(self.contents)

		# Be sure ifp library is cleared & ready for initialization
		self.ifpNotReady = True

# ------------------------------------------------------------------------------------
# Prepare iFP Device
	def ifp(self, parent="",warn=True):

		showWarning = False

		# If device hasn't been connected yet:
		if (self.ifpNotReady):  			
			try:
				# Start up connection
				self.ifpp = None
				self.ifpp = ifp.iFP()
				rtn = self.ifpp.init_device()
				if (self.ifpp == None):
					self.ifpp.get_battery()
				self.ifpNotReady = False
			except:
				showWarning = True

		# Device was previously set up, let's ping it to 
		# make sure it's still connected
		else: 	
			try:
				self.ifpp.get_battery()
			except:
				self.ifpp = None
				self.ifpNotReady = True
				showWarning = True

		# Now, let user know if there was a problem
		if (showWarning):
			print "Couldn't communicate; ",
			if warn:
				self.showError("Unable to find iFP device. Is it plugged in and turned on?\n"+
					"Disconnect and reconnect the USB cable, then retry.\n\n"+
					"Also be sure current user is in 'plugdev' group:\n"+
					"Type  'sudo adduser [name] plugdev'  in a command window,\nlog out, and reconnect.")

		# All done
		return self.ifpp		
				
# ------------------------------------------------------------------------------------
# setup Drag and Drop
	def setupDragAndDrop(self, parent=""):
		self.tree.enable_model_drag_dest([('text/uri-list', 0, 0)],
				  gtk.gdk.ACTION_DEFAULT | gtk.gdk.ACTION_MOVE)
		self.tree.connect("drag-data-received", self.drag_data_received_cb)

		# Experimental - enable drag (for upload to PC)
		self.tree.enable_model_drag_source(gtk.gdk.BUTTON1_MASK,
										  [('STRING',0,0)],
										  gtk.gdk.ACTION_DEFAULT | gtk.gdk.ACTION_MOVE)
		self.tree.connect("drag-data-get", self.drag_data_get_cb)
		self.tree.connect("drag-end", self.dropHappened)

	def drag_data_get_cb(self, treeview, context, selection, info, timestamp):
		treeselection = treeview.get_selection()
		model, iter = treeselection.get_selected()
		text = model.get_value(iter, 0)
		a,b = context.drag_find_window(treeview.get_parent_window(),0,0)
#		print "Chombie!: info:",info,"  time:",timestamp,"  a,b:",a,b

	def dropHappened(self, widget, drag_context):
		a,b = drag_context.drag_find_window(widget.get_parent_window(),0,0)
#		print "drop event!: ",a,b
		

	def drag_data_received_cb(self, treeview, context, x, y, selection, info, timestamp):
		data = selection.data

		# find the location the user dropped on
		model = treeview.get_model()
		drop_info = treeview.get_dest_row_at_pos(x, y)
		drop_path = None

		if drop_info:
			path, position = drop_info
			iter = model.get_iter(path)
			drop_path = self.getFullPath(model,iter)
			# we want a folder for a drop path
			# if this is a file on the ifp, then get the parent
			try:
				if self.ifp().isfile(drop_path):
					drop_path = self.getFullPath(model,model.iter_parent(iter))
			except Exception,e:
				print "Error finding drop folder:",e
			print drop_path
		playlists = []
		folders = []

		lines = data.split('\r\n')
		for line in lines:
			print line
			if (line.startswith("file://")):
				dropfile = urllib.unquote(line[7:])
				print dropfile
				if (dropfile.endswith(".m3u")):
					playlists.append(dropfile)
				else:
					folders.append(dropfile)

		# Now upload them!
		if (len(folders)>0):
			self.uploadFolder(folders,drop_path)
		if (len(playlists)>0):
			self.addPlaylist(playlists)

		return

# ------------------------------------------------------------------------------------
# deleteClicked: Delete the current node and children
	def deleteClicked(self, parent=""):
		tmodel, node = self.tree.get_selection().get_selected() 

		path = self.getFullPath(tmodel,node)
		if ((path == "/") | (path == "")):
			return
		
		try:
			if self.ifp().isdir(path):
				self.ifp().rmdir(path,True)
			else:
				self.ifp().unlink(path)
		except:
			self.showError("""Unable to remove "%s" from iFP device""" % path)

		self.refreshClicked()

		return

# ------------------------------------------------------------------------------------
	def getFullPath(self, model, node, text=""):
		if (node == None):
			return text

		content = "\\" + model.get_value(node,0)
		content = self.getFullPath(
			model, 
			model.iter_parent(node), 
			content+text) 
		return content
		

# ------------------------------------------------------------------------------------
# transferCanceled: cALLBACK to set interrupted flag to cancel transfers
	def cancelTransfer(self, parent=""):
		self.interrupted = True

# ------------------------------------------------------------------------------------
# refreshClicked: Refresh folder view of IFP contents
	def refreshClicked(self, parent=""):
		self.contents = gtk.TreeStore(str,str)
		self.tree.set_model(self.contents)
		while gtk.events_pending(): 
			gtk.main_iteration(False)
		
		self.displayContents()
		self.updateStatusBar()
		self.interrupted = False

		return

# ------------------------------------------------------------------------------------
# addPlaylistClicked: Pick a playlist to upload.
	def addPlaylistClicked(self, parent=""):
		dialog = gtk.FileChooserDialog(title="Select Playlist to Upload",
			parent=self.mainWindow,
			buttons=(gtk.STOCK_CANCEL,gtk.RESPONSE_CANCEL,gtk.STOCK_OPEN,
				gtk.RESPONSE_OK) ) 

		dialog.set_select_multiple(True)
		response = dialog.run()
		filenames = dialog.get_filenames()

		dialog.hide()
		dialog.destroy()		

		while gtk.events_pending(): 
			gtk.main_iteration(False)

		if (response == gtk.RESPONSE_OK):
			self.addPlaylist(filenames)

		return

# ------------------------------------------------------------------------------------
# addPlaylist: Upload a m3u playlist to the player
	def addPlaylist(self, playlists):

		# First calc size of playlists
		self.uploadSize = 0
		names = ""

		for playlist in playlists:
			f = open(playlist)
			for line in f.readlines():
				ff = line[:-1]
				if os.path.exists(ff):
					size = os.path.getsize(line[:-1])
					self.uploadSize += size
			f.close()
			names = names + playlist.split('/')[-1][:-4] + ","
			
		print "** Total size is",self.uploadSize
		names = names[:-1]

		space = self.getFreeSpace()

		# If space < total playlist size, message should
		# warn that full list can't be uploaded.
		message = 'Upload playlist "'+names+'" to iRiver device?'
		cbox = "_Shuffle song order"
		if (space < self.uploadSize):
			message = "The entire playlist will not fit in the current space available.\nTransfer partial playlist?"
			cbox = "Pick _songs from playlist randomly"

		self.verDialog.get_widget("msgText").set_text(message)
		self.verDialog.get_widget("checkShuffle").set_label(cbox)
		self.play = playlists
		self.verifyWindow.show()

# ------------------------------------------------------------------------------------
	def transferCancelled(self, parent=""):
		self.verifyWindow.hide()

# ------------------------------------------------------------------------------------
#   Get free space
	def getFreeSpace(self, parent=""):
		try:
			space = self.ifp().get_free_space()
		except:
			space = -1
		return space

# -----------------------------------------------------
# Calculate entire size of a full playlist
	def calcEntireSize(self, parent=""):
		space = self.getFreeSpace()
		bigsize = 0

		for playlist in self.play:
			# Get list of songs
			try:
				f = open(playlist)
				songs = f.readlines()
				f.close()		

				playlistdir = os.path.dirname(playlist)
			except:
				print "Couldn't read playlist."
				continue

			for song in songs:
				try:
					songpath = os.path.join(playlistdir, song[:-1])
					bigsize += os.path.getsize(songpath)
				except:
					print "Couldn't find "+songpath

		# Can't transfer more than we have room
		if (bigsize > space):
			bigsize = space

		return bigsize

# ------------------------------------------------------------------------------------
#   Upload Playlists to Device
	def transferClicked(self, parent=""):
		shuffle = self.verDialog.get_widget("checkShuffle").get_active()
		missingList = []

		self.verifyWindow.hide()
		self.progBar.set_text("Transferring")
		self.progBar.set_fraction(0.0)

		self.progTree.get_widget("btnProgressCancel").set_sensitive(True)

		space = self.getFreeSpace()

		#Calc size of entire transfer
		self.uploadSize = self.calcEntireSize()
		self.doneSoFarSize = 0

		while gtk.events_pending(): 
			gtk.main_iteration(False)

		for playlist in self.play:
			# Show progress window
			pLabel = self.progTree.get_widget("progLabel")
			pLabel.set_text("Copying "+(playlist.split("/"))[-1] +" to iRiver device")
			self.interrupted=False
			self.progressWindow.show()
			while gtk.events_pending(): 
				gtk.main_iteration(False)

			# Get list of songs
			f = open(playlist)
			songs = f.readlines()
			f.close()		

			# Be sure cancel button wasn't clicked
			if (self.interrupted):
				break;

			# Create playlist folder
			folder = playlist.split("/")[-1]
			if (folder.endswith(".m3u")):
				folder = folder[:-4]

			try:
				self.ifp().mkdir(folder)
			except:
				print ("""Unable to create directory "%s".""" % folder)
				
			# Copy songs to folder
			totalsongs = len(songs)

			while (len(songs)>0):
				song = ""
				if (shuffle):
					# pick a random song
					num = random.randrange(len(songs))
					song = songs[num]
					songs[num:num+1] = []
				else:
					# or pick the next song
					song = songs[0]
					songs[0:1] = []

				# get rid of \n, calculate absolute path, and verify the song exists
				songpath = os.path.join(os.path.dirname(playlist),song[:-1])

				# Make sure file actually exists
				if (False == os.path.exists(songpath)):
					missingList.append(song)
					continue
				size = os.path.getsize(songpath)

				# Make sure there's room
				if (size > space):
					break

				# Be sure cancel button wasn't clicked
				if (self.interrupted):
					break;

				destname = folder+"\\"+songpath.split("/")[-1]

				try:
					self.ifp().upload(songpath,destname,self.uploadStatusCallback)
				except Exception,ee:
					print "Upload failed/cancelled."

				while gtk.events_pending(): 
					gtk.main_iteration(False)

				space -= size
				self.doneSoFarSize += size

		# All done
		self.progBar.set_fraction(1.0)
		self.progressWindow.hide()
		
		if (len(missingList) > 0):
			msg = "Some songs could not be transferred:\n\n"
			for xx in missingList[0:5]:
				msg += xx
			if (len(missingList)>6):
				msg += "...."
				
			self.showError(msg)

		self.refreshClicked()

# ------------------------------------------------------------------------------------
# updateStatusBar: Populate statusbar with battery, size info
	def updateStatusBar(self, parent=""):

		# Get battery status
		try:
			battery = self.ifp(warn=False).get_battery()
			percent = 25 * int(battery)
			message = " Battery: "+ str(percent) + "%"
			self.lblBattery.set_text(message)
		except:
			print "Unable to get battery from the iFP device"
		
		# Get remaining space status
		try:
			cap = self.ifp(warn=False).capacity
			space = self.ifp(warn=False).get_free_space()
			pctused = int(100 * (1.0 - (float(space)/float(cap))))
			message = "Space: "+str(pctused)+"% full "

			#add new msg
			self.lblSpace.set_text(message)
		except:
			print "Unable to get remaining space from the iFP device"
			


# ------------------------------------------------------------------------------------
# addClicked: Add a folder of songs
	def addClicked(self, parent=""):
		dialog = gtk.FileChooserDialog(title="Choose folder to upload",
			parent=self.mainWindow,
			action=gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER,
			buttons=(gtk.STOCK_CANCEL,gtk.RESPONSE_CANCEL,gtk.STOCK_OPEN,
				gtk.RESPONSE_OK) ) 

		dialog.set_select_multiple(True)
		response = dialog.run()
		filenames = dialog.get_filenames()

		dialog.hide()
		dialog.destroy()		

		while gtk.events_pending(): 
			gtk.main_iteration(False)

		if (response == gtk.RESPONSE_OK):
			self.uploadFolder(filenames)

			# Refresh contents!
			self.refreshClicked()
		return


# ------------------------------------------------------------------------------------
# displayContents: Pull the current iRiver contents off of the player
# and populate the treeview - recursively.
	def displayContents(self, parent=None, pDir=""):
		path = self.getFullPath(self.tree.get_model(),parent,"")
		try:
			output = self.ifp().listdir(path)
			if output is not None:
				for line in output:
					if line[2] == ifp.IFP_DIR:
						dirName = line[0]
						node = self.contents.append(parent,(dirName,""))
						self.displayContents(node,dirName)
					elif line[2] == ifp.IFP_FILE:
						filesize = line[1] / 1024
						filesize = self.commafy(filesize)+"k"
						node = self.contents.append(parent,
							(line[0],filesize))
		except:
			print "Got exception."
			pass
		self.tree.collapse_all()
		return		

# ------------------------------------------------------------------------------------
# Calc tree node size
	def getNodeSize(self, tmodel,node):
		size = self.uncommafy(tmodel.get_value(node,1))
		kid = tmodel.iter_children(node)
		while kid:
			size += self.getNodeSize(tmodel,kid)
			kid = tmodel.iter_next(kid)
		return size
		
# ------------------------------------------------------------------------------------
# Copy to PC
	def copyToPC(self, parent=""):

		# Figure out current selection
		tmodel, node = self.tree.get_selection().get_selected() 
		path = self.getFullPath(tmodel,node)
		name = tmodel.get_value(node,0)

		if ((path == "/") | (path == "")):
			return
		
		# Pick where to upload (maybe replace this later when DnD works
		dialog = gtk.FileChooserDialog(title="Select Destination Folder",
			parent=self.mainWindow,
			action=gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER,
			buttons=(gtk.STOCK_CANCEL,gtk.RESPONSE_CANCEL,gtk.STOCK_OPEN,
				gtk.RESPONSE_OK) ) 

		dialog.set_select_multiple(False)
		response = dialog.run()
		folders = dialog.get_filenames()
		
		dialog.hide()
		dialog.destroy()		

		while gtk.events_pending(): 
			gtk.main_iteration(False)

		if (response == gtk.RESPONSE_OK):
			# TODO: Check if already exists (and OK to replace)

			# Calculate upload size
			self.uploadSize = self.getNodeSize(tmodel,node) * 1024
			self.doneSoFarSize = 0
			self.interrupted=False

			# Show progress window
			pLabel = self.progTree.get_widget("progLabel")
			pLabel.set_text("Copying "+name+" from device...")
			self.progressWindow.show()

			try:
				for folder in folders:
					x = folder+"/"+name
					self.ifp().download(path,x,self.uploadStatusCallback)
			except Exception,e:
				msg = str(e)
				print msg
				if (msg=="Unknown error occured"):
					msg +="\nDoes folder already exist?"
				self.showError(msg)

			self.progressWindow.hide()



# ------------------------------------------------------------------------------------
# uploadStatusCallback
	def uploadStatusCallback(self,transfer_status):
		thisTransferSize = 0

		if transfer_status.is_batch:
			thisTransferSize = transfer_status.batch_bytes #) / transfer_status.batch_total;
		else:
			thisTransferSize = transfer_status.file_bytes
			#fraction = (1.0 * transfer_status.file_bytes) / transfer_status.file_total;

		fraction = (1.0 * self.doneSoFarSize + thisTransferSize) / self.uploadSize

		self.progBar.set_fraction(fraction)
		self.progBar.set_text(str(int(100 * fraction)) + "% complete")

		while gtk.events_pending(): 
			gtk.main_iteration(False)

		return self.interrupted
		
# ------------------------------------------------------------------------------------
# uploadFolder
	def uploadFolder(self, folders,target=None):
		self.uploadSize = 0
		itemSize = []
		
		# Calculate full size of upload
		for item in folders:

			if (os.path.isdir(item)):
				size = 0
				for root, dirs, files in os.walk(item):
					size += sum([os.path.getsize(os.path.join(root,name)) for name in files])
			else:
				size = os.path.getsize(item)

			itemSize.append(size)
			self.uploadSize += size

		if (self.uploadSize == 0):
			self.uploadSize=1  #to prevent div-by-zero

		print "total size: ",self.uploadSize

		self.doneSoFarSize = 0

		#Test available space on iRiver
		space = self.getFreeSpace()
		if (self.uploadSize > space):
			# Files are too big; for now just send an error
			# TODO: ask if a subset should be downloaded
			errmsg = "There is not enough room on the device for those files."
			self.showError(errmsg)
			return				
		
		#Put folder onto iRiver device
		progressItem = 0
		for folder in folders:
			# Show progress window
			pLabel = self.progTree.get_widget("progLabel")
			foldername = folder.split("/")
			pLabel.set_text("Copying "+foldername[-1]+" to device...")
			self.interrupted=False

			self.progressWindow.show()
			print "folder = %s" % folder

			# Determine target folder
			dest = foldername[-1]
			if (target != None):
				dest = target + "\\" + dest

			# Be sure cancel button wasn't clicked
			if (self.interrupted):
				break;

			try:
				self.ifp().upload(folder, dest, self.uploadStatusCallback)
			except Exception,e:
				print "Got exception on upload:",e
			
			self.doneSoFarSize += itemSize[progressItem]
			progressItem += 1
			  
		self.progressWindow.hide()
		self.refreshClicked()
		return

# ------------------------------------------------------------------------------------
# showError: Show an error dialog with the supplied message and return the dialog's return code
	def showError(self,message = ""):
		errmsg = "Error: "
		errmsg += message
		bad = gtk.MessageDialog(self.mainWindow, gtk.DIALOG_MODAL,
			gtk.MESSAGE_ERROR, gtk.BUTTONS_OK, errmsg)
		response = bad.run()
		bad.destroy()
		return response

# ----------------------------------------------
	def commafy(self,val):
	     s = str(val)
	     start = len(s) % 3

	     chunks = []
	     if start > 0:
	         chunks.append(s[:start])
	
	     for i in range(start, len(s), 3):
	         chunks.append(s[i:i+3])
	
	     return ','.join(chunks)

# ----------------------------------------------
	def uncommafy(self,st):
		 if (st == None or st == ""):
			return 0

		 st = st.replace(",","")
		 st = st.replace("k","")
		 return int(st)

# ----------------------------------------------
	def showAboutDialog(self, parent=""):
		gnome.ui.About("iFP-Gnome","0.7", "Copyright 2006 Billy Charlton",
             "A GNOME file manager for some iRiver Flash devices",
             ["Billy Charlton <billy@trigby.com>"],[],"",
             gtk.gdk.pixbuf_new_from_file("ifp-gnome.png")).show()


# MAIN ENTRY POINT ---------------------

app = GiRiverApp()

#Show window
app.wTree.get_widget("window1").show()
while gtk.events_pending(): 
	gtk.main_iteration(False)

#Set up iFP device
app.ifp(warn=False)

#Show contents
app.displayContents()
app.updateStatusBar()
gtk.main()
